from flask_restx import fields, Model

bearer_token_model = Model("Token", {
    "token": fields.String(
        description="Bearer token",
        pattern="Bearer [a-zA-Z_\\d]{100}"),
    "expiresIn": fields.String(
        description="Token expiration date"
    )
})



from flask_sqlalchemy import SQLAlchemy
from ..config import Config
from ..utils import info, EXTENSION_BIND

db = SQLAlchemy()


def init_app(config_object, testing: bool, declarative_base: SQLAlchemy):

    # Simulate schema in sqlite
    sqlite_schema(declarative_base)

    # Create all necessary database entities
    if (hasattr(config_object, "DB_CREATE_ENTITIES") and bool(config_object.DB_CREATE_ENTITIES)) or testing:
        declarative_base.create_all(bind_key=EXTENSION_BIND)
        info("Database Model Initialized")


def sqlite_schema(declarative_base: SQLAlchemy):

    # Simulate schema in sqlite
    engine = declarative_base.engines[EXTENSION_BIND]
    if engine.url.get_backend_name() == "sqlite":
        connection = engine.raw_connection()
        try:
            # Get a SQLite cursor from the connection
            cursor = connection.cursor()
            # Execute the ATTACH DATABASE command
            cursor.execute(f"attach database ':memory:' as '{Config.DB_SCHEMA}'")
            # Commit the transaction
            connection.commit()
        finally:
            # Close the connection
            connection.close()



from flask import json, Response
from functools import wraps
from http import HTTPStatus
from typing import Callable

from ..utils.utils import error
from ..exceptions import HttpResponseError


def response_api_ok(http_code, data: dict, headers: dict = {}, endpoint: str = None) -> Response:

    data = {**data, "status": "OK", "code": http_code}
    if endpoint is not None:
        data["endpoint"] = endpoint
    response = Response(
        response=json.dumps(data),
        status=http_code,
        mimetype="application/json",
        headers=headers
    )
    return response


def response_api_error(http_code, error, headers: dict = {}, endpoint: str = None) -> Response:

    data = {"error": error, "status": "ERROR", "code": http_code}
    if endpoint is not None:
        data["endpoint"] = endpoint
    response = Response(
        response=json.dumps(data),
        status=http_code,
        mimetype="application/json",
        headers=headers
    )
    return response


def handle_exceptions(**kwargs) -> Callable:

    endpoint: str = kwargs.get("endpoint", None)

    def wrapper(route_function):

        @wraps(route_function)
        def decorated_function(*args, **kwargs):
            endpoint_id = route_function.__qualname__ if not endpoint else endpoint
            try:
                #
                # return current_app.ensure_sync(route_function)(*args, **kwargs)
                return route_function(*args, **kwargs)
            except HttpResponseError as e:
                error_msg = f"Bearer {e.error_description()}"
                error(f"{error_msg}")
                http_code = e.status_code
                headers = {"WWW-Authenticate": f"{error_msg}"}
                return response_api_error(http_code=http_code, error=error_msg, headers=headers, endpoint=endpoint_id)
            except Exception as e:
                error_msg = f"{e}"
                http_code = int(HTTPStatus.INTERNAL_SERVER_ERROR.value)
                headers = {"WWW-Authenticate": f"{error_msg}"}
                return response_api_error(http_code=http_code, error=error_msg, headers=headers, endpoint=endpoint_id)
            except BaseException as e:
                error_token = "Token validation failed"
                error_msg = f"Bearer error=\"invalid_token\" error_description=\"{error_token}\""
                error(f"{error_msg}. Exception: {e}")
                http_code = int(HTTPStatus.UNAUTHORIZED.value)
                headers = {"WWW-Authenticate": f"{error_msg}"}
                return response_api_error(http_code=http_code, error=error_msg, headers=headers, endpoint=endpoint_id)

        decorated_function.__name__ = route_function.__name__
        return decorated_function
    return wrapper



import base64
from .logger import logger


def string_base64(s):
    return base64.b64encode(s.encode("utf-8")) if isinstance(s, str) else None


def base64_string(b):
    return base64.b64decode(b).decode("utf-8") if isinstance(b, bytes) else None


# Shortcut to logger functions
def info(message, *args, **kwargs):
    logger.info(message, *args, **kwargs)


def debug(message, *args, **kwargs):
    logger.debug(message, *args, **kwargs)


def error(message, *args, **kwargs):
    logger.error(message, *args, **kwargs)


def warning(message, *args, **kwargs):
    logger.warning(message, *args, **kwargs)


def critical(message, *args, **kwargs):
    logger.critical(message, *args, **kwargs)



import base64

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import \
    RSAPrivateKeyWithSerialization as RSAPrivateKey
from cryptography.hazmat.primitives.asymmetric.rsa import \
    RSAPublicKeyWithSerialization as RSAPublicKey
from cryptography.hazmat.backends import default_backend

from collections import namedtuple

RSAKeyTuple = namedtuple("RSAKeyTuple", ("private", "public"))


def generate_rsa(bits: int = 2048) -> RSAKeyTuple:
    """
    Generate an RSA tuple in PEM format
    :param bits:
        Number of bits to create the key
    :return:
        Private key and public key
    """

    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=bits,
        backend=default_backend(),
    )

    public_key = private_key.public_key()

    return RSAKeyTuple(private=private_key, public=public_key)


def decode_rsa_private(private_key: RSAPrivateKey, pem_format=True) -> str:
    if pem_format:
        return private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            encryption_algorithm=serialization.NoEncryption(),
            format=serialization.PrivateFormat.PKCS8
        ).decode()
    else:
        return base64.b64encode(
            private_key.private_bytes(
                encoding=serialization.Encoding.DER,
                encryption_algorithm=serialization.NoEncryption(),
                format=serialization.PrivateFormat.PKCS8
            )
        ).decode()


def decode_rsa_public(public_key: RSAPublicKey, pem_format=True) -> str:
    if pem_format:
        return public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.PKCS1
        ).decode()
    else:
        return base64.b64encode(
            public_key.public_bytes(
                encoding=serialization.Encoding.DER,
                format=serialization.PublicFormat.PKCS1
            )
        ).decode()


def encode_private_key(data: str) -> RSAPrivateKey:
    return serialization.load_der_private_key(
        data=base64.b64decode(data),
        password=None,
        backend=default_backend()
    )


def encode_public_key(data: str) -> RSAPublicKey:
    return serialization.load_der_public_key(
        data=base64.b64decode(data),
        backend=default_backend()
    )



import sys
from http import HTTPStatus


class OceanaError(Exception):
    """
    Base exception for all errors.

    :param object message: The message object stringified as 'message' attribute
    :keyword error: The original exception if any
    :paramtype error: Exception

    :ivar inner_exception: The exception passed with the 'error' kwarg
    :vartype inner_exception: Exception
    :ivar exc_type: The exc_type from sys.exc_info()
    :ivar exc_value: The exc_value from sys.exc_info()
    :ivar exc_traceback: The exc_traceback from sys.exc_info()
    :ivar exc_msg: A string formatting of message parameter, exc_type and exc_value
    :ivar str message: A stringified version of the message parameter
    """

    def __init__(self, message, *args, **kwargs) -> None:
        self.inner_exception = kwargs.get("error")

        exc_info = sys.exc_info()
        self.exc_type = exc_info[0]
        self.exc_value = exc_info[1]
        self.exc_traceback = exc_info[2]

        self.exc_type = self.exc_type if self.exc_type else type(self.inner_exception)
        self.exc_msg: str = "{}, {}: {}".format(message, self.exc_type.__name__, self.exc_value)
        self.message: str = str(message)
        super().__init__(self.message, *args)


class ServiceRequestError(OceanaError):
    """
    An error occurred while attempt to make a request to the service. (with status code 400)
    No request was sent.
    """


class ServiceResponseError(OceanaError):
    """
    The request was sent, but the client failed to understand the response.
    The connection may have timed out. These errors can be retried for idempotent or
    safe operations
    """


class ServiceRequestTimeoutError(ServiceRequestError):
    """
    Error raised when timeout happens
    """


class ServiceResponseTimeoutError(ServiceResponseError):
    """
    Error raised when timeout happens
    """


class HttpResponseError(OceanaError):
    """
    A request was made, and a non-success status code was received from the service.

    :param object message: The message object stringified as 'message' attribute
    :param response: The response that triggered the exception.
    """

    # def __init__(self, message=None, response=None, **kwargs) -> None:
    def __init__(self,
                 status_code: int,
                 error: str,
                 message: str,
                 response=None,
                 **kwargs) -> None:

        self.status_code = status_code
        self.message = message
        self.error = error
        self.response = response
        self.reason = kwargs.get("reason")

        if response:
            self.reason = response.reason
            self.status_code = response.status_code

        super().__init__(message=message,
                         error=error,
                         **kwargs)

    def error_description(self):
        """
        Error description to add to header `WWW-Authenticate`
        """
        return f"error=\"{self.error}\" error_description=\"{self.message}\""


# 401 UNAUTHORIZED
class ClientAuthenticationError(HttpResponseError):
    """
    An client error response code 401 UNAUTHORIZED
    """
    def __init__(self, message, error="invalid_token", response=None, **kwargs) -> None:
        super().__init__(status_code=int(HTTPStatus.UNAUTHORIZED.value),
                         error=error,
                         message=message,
                         response=response,
                         **kwargs)


# Oauth2 errors
class ClientJWTDecodeException(ClientAuthenticationError):
    """
    Decode error response code 401 UNAUTHORIZED
    """
    def __init__(self, message, response=None, **kwargs) -> None:
        super().__init__(message=message,
                         error="invalid_token",
                         response=response, **kwargs)


class ClientIssuerException(ClientAuthenticationError):
    """
    Invalid issuer error response code 401 UNAUTHORIZED
    """
    def __init__(self, message, response=None, **kwargs) -> None:
        super().__init__(message=message,
                         error="invalid_token",
                         response=response, **kwargs)


# 400 BAD_REQUEST
class ClientBadRequestException(HttpResponseError):
    """
    An bad request error with response code 400 BAD_REQUEST
    """
    def __init__(self, message, response=None, **kwargs) -> None:
        super().__init__(status_code=int(HTTPStatus.BAD_REQUEST.value),
                         error="invalid_request",
                         message=message,
                         response=response,
                         **kwargs)


class ClientInvalidSignatureException(ClientBadRequestException):
    """
    Invalid signature error response code 400 BAD_REQUEST
    """
    def __init__(self, message, response=None, **kwargs) -> None:
        super().__init__(message=message,
                         response=response, **kwargs)



from enum import Enum

# Flask extension name
EXTENSION_NAME = "oceana-jwt-auth"

# Bind for SQLAlchemy
EXTENSION_BIND = "oceana_jwt_auth"

ENDPOINT_SECURITY_LABEL = "ENDPOINT_SECURITY"

# Authorization api version (in URL)
API_AUTH_DESCRIPTION = "Authentication API"


# Default values for Flask RestX Api
API_AUTH_DEFAULT_TITLE = "Test API"
API_AUTH_DEFAULT_VERSION = "1.0"
API_AUTH_DEFAULT_DESCRIPTION = "Test API"


class RestMethod(Enum):
    GET = "GET"
    POST = "POST"
    DELETE = "DELETE"
    PUT = "PUT"
    PATCH = "PATCH"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    TRACE = "TRACE"
    CONNECT = "CONNECT"


# Security Authorization
class AuthClientType(Enum):
    APPLICATION = "application"
    USER = "user"


class AuthAPIRoles(Enum):
    ALLOW_ALL = "all"
    ADMIN = "admin"
    READER = "reader"
    WRITER = "writer"



from typing import Dict

from .utils.constants import AuthClientType
from .utils.utils import debug
from .utils.crypto import hash_string, safe_str_cmp
from .database import get_identity
from .exceptions import ClientAuthenticationError


def authenticate(provider, client_id, client_secret) -> Dict:

    identity = get_identity(
        provider=provider,
        client_type=AuthClientType.APPLICATION.value,  # Only applications
        client_id=client_id)

    if identity is not None:
        db_hashed, client_salt, roles = \
            identity[0].client_hash, \
            identity[0].client_salt, \
            identity[0].roles,

        secret_hashed = hash_string(source_str=client_secret, salt=client_salt)

        # Check password against hash
        if not safe_str_cmp(secret_hashed, db_hashed):
            raise ClientAuthenticationError(f"Invalid credentials for client id: \"{client_id}\"")

        # Create list of roles from database
        role_list = [r.strip().lower() for r in str(roles).strip().split(",")]
        debug(f"Client id: {client_id} authenticated OK. Roles: {role_list}")

        # Return identity
        return {
            "client_id": client_id,
            "client_type": AuthClientType.APPLICATION.value,
            "roles": role_list
        }
    else:
        # User not found in the repository, same message to avoid give more information
        raise ClientAuthenticationError(f"Invalid credentials for client id: \"{client_id}\"")



from typing import TYPE_CHECKING
from typing import List, Any
from flask import current_app

from .utils import EXTENSION_NAME
from .utils.utils import error, debug
from .utils.constants import AuthAPIRoles
from .config import config
from .exceptions import ClientAuthenticationError

# from flask.typing import ResponseReturnValue

if TYPE_CHECKING:  # pragma: no cover
    from .jwt_extension import JWTExtension


def get_jwt_extension() -> "JWTExtension":
    try:
        return current_app.extensions[EXTENSION_NAME]
    except KeyError:
        raise RuntimeError(
            "JWTExtension must be initialized in application before using this method"
        ) from None


def is_authorized(roles, allowed) -> bool:

    # Role all_users gran access to api endpoint
    if AuthAPIRoles.ALLOW_ALL.value in allowed:
        return True

    # Admin can access to everything
    if AuthAPIRoles.ADMIN.value in roles:
        return True
    authorized = False
    for r in roles:
        authorized = r in allowed
        if authorized:
            break
    return authorized


def default_token_verification_callback(endpoint_id: str,
                                        jwt_header: dict,
                                        jwt_data: dict,
                                        optional: bool,
                                        allowed: List[str],
                                        roles: List[str]) -> None:
    """
    Default token verification

    :param str endpoint_id: Argument indicating the endpoint identificator.
    :param dict jwt_header: Dictionary containing the header data of the JWT.
    :param dict jwt_data: Dictionary containing the payload data of the JWT.
    :param bool optional: Boolean ``True`` if JWT is optional in the endpoint
        ``False`` otherwise.
    :param list allowed: List of roles allowed for the endpoint.
    :param list roles: List of roles from the valid JWT

    :return None:
        Throw HttpResponseError exception if not allowed.
    """
    # TODO: verification of the user claims.
    # error_msg = "User claims verification failed"
    # raise UserClaimsVerificationError(error_msg, jwt_header, jwt_data)

    # Authorization gate
    client_id = jwt_data.get(config.identity_claim_key, "unknown") if len(jwt_data) > 0 else "unknown"

    if config.api_secured and not optional:
        assert jwt_header["typ"] == "JWT"
        assert jwt_header["alg"] == config.algorithm

        allow = is_authorized(roles=roles, allowed=allowed)
        if not allow:
            error(f"Endpoint \"{endpoint_id}\". Authorization not granted for client " +
                  f"\"{client_id}\", roles: {roles}, allowed: {allowed}")
            error_msg = "Authorization required"
            raise ClientAuthenticationError(error_msg)

    debug(f"Endpoint \"{endpoint_id}\". Authorization for client_id: \"{client_id}\"")


def default_user_claims_callback(identity: Any) -> dict:
    return {}


def default_encode_key_callback(identity: Any, algorithm) -> str:

    key = config.rsa_private_key if config.requires_cryptography \
        else config.secret_key
    if not key:
        raise RuntimeError("Can't encode without private key")
    return key


def default_decode_key_callback(identity: Any, algorithm) -> str:
    key = config.rsa_public_key if config.requires_cryptography \
        else config.secret_key
    if not key:
        raise RuntimeError("Can't decode without public key")
    return key


def default_token_header_callback(headers: dict) -> dict:
    """
    Use this callback to add labels to JWT tokens header.
    We don't add more labels by default. The two properties of JWT
    header will be set automatically when encoding the payload.
    These properties are the encoding algorithm and the type of token.
    {'alg': 'HS256', 'typ': 'JWT'}

    :return: Dictionary with properties to add
    """
    return {}



from decouple import config as _config
from flask import current_app
from urllib.parse import quote_plus
from jwt.algorithms import requires_cryptography, get_default_algorithms
from datetime import timedelta

from typing import Union, Iterable

# Database provider and issuer of JWT tokens
OCEANA_API_PROVIDER = _config("OCEANA_API_PROVIDER", default="OceanaAPI", cast=str)

# Security properties
OCEANA_API_SECURED = _config("OCEANA_API_SECURED", default=True, cast=bool)
# Oceana API secret key
OCEANA_API_SECRET_KEY = _config("OCEANA_API_SECRET_KEY", default="", cast=str)
# Generate a token with valid within 1 hour by now (in minutes)
OCEANA_API_TOKEN_MAX_MINUTES = _config("OCEANA_API_TOKEN_MAX_MINUTES", default=60, cast=int)

# Encode and decode issuer in JWT tokens
OCEANA_API_TOKEN_ENCODE_ISSUER = _config("OCEANA_API_TOKEN_ENCODE_ISSUER", default=OCEANA_API_PROVIDER)
OCEANA_API_TOKEN_DECODE_ISSUER = _config("OCEANA_API_TOKEN_DECODE_ISSUER", default=OCEANA_API_PROVIDER)

# RSA algorithm private and public keys
OCEANA_API_RSA_PRIVATE_KEY = _config("OCEANA_API_RSA_PRIVATE_KEY", default="", cast=str)
OCEANA_API_RSA_PUBLIC_KEY = _config("OCEANA_API_RSA_PUBLIC_KEY", default="", cast=str)

# Logger Configuration
OCEANA_API_LOGGING_LEVEL = _config("OCEANA_API_LOGGING_LEVEL", "INFO")
OCEANA_API_LOGGING_DIR = _config("OCEANA_API_LOGGING_DIR", "logs")
OCEANA_API_LOGGING_FILE = _config("OCEANA_API_LOGGING_FILE", "oceana-jwt-auth.log")
OCEANA_API_LOGGING_WHEN = _config("OCEANA_API_LOGGING_WHEN", "midnight")
OCEANA_API_LOGGING_INTERVAL = _config("OCEANA_API_LOGGING_INTERVAL", 1)
OCEANA_API_LOGGING_TITLE = _config("OCEANA_API_LOGGING_TITLE", "oceana-jwt-auth")

OCEANA_API_LOGGING_HANDLERS = _config("OCEANA_API_LOGGING_HANDLERS", "console,file_handler")
OCEANA_API_LOGGING_FORMATTER = _config("OCEANA_API_LOGGING_FORMATTER",
                                       "%(asctime)s - [%(name)s] - %(levelname)-5s - %(message)s")

OCEANA_API_TOKEN_ALGORITHM = _config("OCEANA_API_TOKEN_ALGORITHM", "HS256")
# Refresh token expiration in minutes
OCEANA_API_TOKEN_REFRESH_MAX_MINUTES = _config("OCEANA_API_TOKEN_REFRESH_MAX_MINUTES", default=60, cast=int)

# Authorization api version (in URL)
OCEANA_API_AUTH_VERSION = _config("OCEANA_API_AUTH_VERSION", None)  # "v1"

# Audiences
OCEANA_API_TOKEN_ENCODE_AUDIENCE = _config("OCEANA_API_TOKEN_ENCODE_AUDIENCE", None)
OCEANA_API_TOKEN_DECODE_AUDIENCE = _config("OCEANA_API_TOKEN_DECODE_AUDIENCE", None)


# Token version (internal in JWT)
OCEANA_API_TOKEN_VERSION = "v1"

# Verifications
OCEANA_API_TOKEN_VERIFY_SUB = _config("OCEANA_API_TOKEN_VERIFY_SUB", default=True, cast=bool)
OCEANA_API_TOKEN_VERIFY_VERSION = _config("OCEANA_API_TOKEN_VERIFY_VERSION", default=True, cast=bool)

# Register authorization namespace
OCEANA_API_REGISTER_AUTH = _config("OCEANA_API_REGISTER_AUTH", True, cast=bool)

# Token JWT Algorithms
ALGORITHMS = [a for a in get_default_algorithms().keys() if a != "none"]


# Flask application base config
class BaseConfig:

    @property
    def api_secured(self) -> bool:
        return self._get_bool("SECURED", OCEANA_API_SECURED)

    @property
    def secret_key(self) -> str:
        return current_app.config.get("SECRET_KEY") or OCEANA_API_SECRET_KEY

    # Algorithms:
    # HS256 - HMAC using SHA-256 hash algorithm (default)
    # HS384 - HMAC using SHA-384 hash algorithm
    # HS512 - HMAC using SHA-512 hash algorithm
    # ES256 - ECDSA signature algorithm using SHA-256 hash algorithm
    # ES256K - ECDSA signature algorithm with secp256k1 curve using SHA-256 hash algorithm
    # ES384 - ECDSA signature algorithm using SHA-384 hash algorithm
    # ES512 - ECDSA signature algorithm using SHA-512 hash algorithm
    # RS256 - RSASSA-PKCS1-v1_5 signature algorithm using SHA-256 hash algorithm
    # RS384 - RSASSA-PKCS1-v1_5 signature algorithm using SHA-384 hash algorithm
    # RS512 - RSASSA-PKCS1-v1_5 signature algorithm using SHA-512 hash algorithm
    # PS256 - RSASSA-PSS signature using SHA-256 and MGF1 padding with SHA-256
    # PS384 - RSASSA-PSS signature using SHA-384 and MGF1 padding with SHA-384
    # PS512 - RSASSA-PSS signature using SHA-512 and MGF1 padding with SHA-512
    # EdDSA - Both Ed25519 signature using SHA-512 and Ed448 signature using SHA-3 are supported.
    #     Ed25519 and Ed448 provide 128-bit and 224-bit security respectively.

    @property
    def algorithm(self) -> str:
        return current_app.config.get("TOKEN_ALGORITHM") or OCEANA_API_TOKEN_ALGORITHM

    # RSA algorithm
    @property
    def requires_cryptography(self) -> bool:
        return self.algorithm in requires_cryptography

    @property
    def rsa_private_key(self) -> str:
        return current_app.config.get("RSA_PRIVATE_KEY") or OCEANA_API_RSA_PRIVATE_KEY

    @property
    def rsa_public_key(self) -> str:
        return current_app.config.get("RSA_PUBLIC_KEY") or OCEANA_API_RSA_PUBLIC_KEY

    @property
    def access_token_delta(self) -> timedelta:
        delta = current_app.config.get("TOKEN_MAX_MINUTES") or OCEANA_API_TOKEN_MAX_MINUTES
        return timedelta(minutes=delta)

    # TODO: Implement refresh tokens
    # @property
    # def refresh_token_delta(self) -> timedelta:
    #     delta = current_app.config.get("TOKEN_REFRESH_MAX_MINUTES") or OCEANA_API_TOKEN_REFRESH_MAX_MINUTES
    #     return timedelta(minutes=delta)

    # @property
    # def api_auth_version(self):
    #     return current_app.config.get("API_AUTH_VERSION") or OCEANA_API_AUTH_VERSION

    @property
    def encode_issuer(self) -> str:
        return current_app.config.get("TOKEN_ENCODE_ISSUER") or OCEANA_API_TOKEN_ENCODE_ISSUER

    @property
    def decode_issuer(self) -> str:
        return current_app.config.get("TOKEN_DECODE_ISSUER") or OCEANA_API_TOKEN_DECODE_ISSUER

    @property
    def encode_audience(self) -> Union[str, Iterable[str]]:
        return current_app.config.get("TOKEN_ENCODE_AUDIENCE") or OCEANA_API_TOKEN_ENCODE_AUDIENCE

    @property
    def decode_audience(self) -> Union[str, Iterable[str]]:
        return current_app.config.get("TOKEN_DECODE_AUDIENCE") or OCEANA_API_TOKEN_DECODE_AUDIENCE

    @property
    def identity_claim_key(self) -> str:
        return "sub"

    # Verifications
    @property
    def verify_sub(self) -> bool:
        return self._get_bool("TOKEN_VERIFY_SUB", OCEANA_API_TOKEN_VERIFY_SUB)

    @property
    def verify_version(self) -> bool:
        return self._get_bool("TOKEN_VERIFY_VERSION", OCEANA_API_TOKEN_VERIFY_VERSION)

    @property
    def register_auth(self) -> bool:
        return self._get_bool("REGISTER_AUTH", OCEANA_API_REGISTER_AUTH)

    @property
    def token_version(self):
        return OCEANA_API_TOKEN_VERSION

    def _get_bool(self, key: str, env_val) -> bool:
        if key in current_app.config:
            return bool(current_app.config.get(key))
        else:
            return bool(env_val)

    SWAGGER_UI_DOC_EXPANSION = "list"  # ("none", "list" or "full")

    # Default options, if you want to use other parameters in SqlAlchemy
    # use ConfigSqlAlchemy configuration instead of BaseConfig
    SQLALCHEMY_DATABASE_URI = "sqlite:///:memory:"
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SQLALCHEMY_RECORD_QUERIES = False
    DB_SCHEMA = _config("DB_SCHEMA", default="public")

    OCEANA_API_DB_AUTH_PASSWORD = _config("OCEANA_API_DB_AUTH_PASSWORD", default=None)
    OCEANA_API_DB_AUTH_SCHEMA = _config("OCEANA_API_DB_AUTH_SCHEMA", default=DB_SCHEMA)

    # OAUTH2_PROVIDERS = {
    #     "azure": {
    #         "client_id": _config("AZURE_CLIENT_ID", ""),
    #         "client_secret": _config("AZURE_CLIENT_SECRET", ""),
    #         "authorize_url": _config("AZURE_AUTH_URL", ""),
    #         "access_token_url": _config("AZURE_TOKEN_URL", ""),
    #         "client_kwargs": {"scope": "openid email profile"},
    #         "jwks_uri": _config("AZURE_JWKS_URL", ""),

    #     },
    #     "google": {
    #         "client_id": _config("GOOGLE_CLIENT_ID", ""),
    #         "client_secret": _config("GOOGLE_CLIENT_SECRET", ""),
    #         "server_metadata_url": _config("GOOGLE_METADATA_URL", ""),
    #         "client_kwargs": {"scope": "openid email profile"}
    #     },
    #     "twitter": {
    #         "api_base_url": _config("OAUTH2_TWITTER_API_BASE_URL", ""),
    #         "request_token_url": _config("OAUTH2_TWITTER_REQUEST_TOKEN_URL", ""),
    #         "access_token_url": _config("OAUTH2_TWITTER_ACCESS_TOKEN_URL", ""),
    #         "authorize_url": _config("OAUTH2_TWITTER_AUTHORIZE_URL", ""),
    #         "userinfo_endpoint": _config("OAUTH2_TWITTER_USERINFO_ENDPOINT", ""),
    #         "userinfo_compliance_fix": _config("OAUTH2_TWITTER_USERINFO_COMPLIANCE_FIX", "")
    #     }
    # }


# Database Configuration POSTGRESQL
class Config(BaseConfig):
    DB_HOST = _config("DB_HOST", default=None)
    DB_NAME = _config("DB_NAME", default=None)
    DB_USERNAME = _config("DB_USERNAME", default=None)
    DB_PASSWORD = None if (_passwd := _config("DB_PASSWORD", default=None)) is None else quote_plus(_passwd)
    DB_PORT = _config("DB_PORT", default=5432, cast=int)
    DB_SCHEMA = _config("DB_SCHEMA", default="public")
    DB_CREATE_ENTITIES = _config("DB_CREATE_ENTITIES", default=True, cast=bool)


class ConfigSqlAlchemy(Config):
    # SQLAlchemy Configuration
    SQLALCHEMY_DATABASE_URI = _config("SQLALCHEMY_DATABASE_URI", None)
    SQLALCHEMY_ENGINE_OPTIONS = _config("SQLALCHEMY_ENGINE_OPTIONS", {})
    SQLALCHEMY_ECHO = _config("SQLALCHEMY_ECHO", None)
    SQLALCHEMY_BINDS = _config("SQLALCHEMY_BINDS", {})
    SQLALCHEMY_RECORD_QUERIES = _config("SQLALCHEMY_RECORD_QUERIES", None)
    SQLALCHEMY_TRACK_MODIFICATIONS = _config("SQLALCHEMY_TRACK_MODIFICATIONS", None)


# SQLAlchemy Postgres Configuration (overwrite properties)
class ConfigPostgres(ConfigSqlAlchemy):

    SQLALCHEMY_DATABASE_URI = ConfigSqlAlchemy.SQLALCHEMY_DATABASE_URI \
        if ConfigSqlAlchemy.SQLALCHEMY_DATABASE_URI is not None else \
        f"postgresql://{Config.DB_USERNAME}:{Config.DB_PASSWORD}" \
        f"@{Config.DB_HOST}:{Config.DB_PORT}/{Config.DB_NAME}"
    SQLALCHEMY_ECHO = bool(ConfigSqlAlchemy.SQLALCHEMY_ECHO
                           if ConfigSqlAlchemy.SQLALCHEMY_ECHO is not None else False)
    SQLALCHEMY_TRACK_MODIFICATIONS = bool(ConfigSqlAlchemy.SQLALCHEMY_TRACK_MODIFICATIONS
                                          if ConfigSqlAlchemy.SQLALCHEMY_TRACK_MODIFICATIONS
                                          is not None else False)
    SQLALCHEMY_RECORD_QUERIES = bool(ConfigSqlAlchemy.SQLALCHEMY_RECORD_QUERIES
                                     if ConfigSqlAlchemy.SQLALCHEMY_RECORD_QUERIES
                                     is not None else False)


# SQLAlchemy SQLite Configuration (overwrite properties)
# Used in tests
class ConfigSqlite(ConfigSqlAlchemy):

    # SQLALCHEMY_DATABASE_URI = "sqlite:///project.db"
    SQLALCHEMY_DATABASE_URI = "sqlite:///:memory:"
    DB_HOST = None
    DB_NAME = None
    DB_USERNAME = None
    DB_PASSWORD = None
    DB_PORT = None
    # Default value to test
    DB_SCHEMA = "public"
    DB_CREATE_ENTITIES = True


# Syntax sugar: Shared instance of Config class just to get the
# properties (@property label) from the Flask current_app.
config = Config()



from .logger import logger
from .constants import \
    API_AUTH_DESCRIPTION, API_AUTH_DEFAULT_TITLE, \
    API_AUTH_DEFAULT_VERSION, API_AUTH_DEFAULT_DESCRIPTION, \
    RestMethod, AuthClientType, AuthAPIRoles, \
    EXTENSION_NAME, EXTENSION_BIND, ENDPOINT_SECURITY_LABEL
from .utils import string_base64, base64_string, info, debug, error, warning, critical
from .crypto import generate_key, generate_salt, hash_string, safe_str_cmp


__all__ = [
    "logger",
    "API_AUTH_DESCRIPTION",
    "API_AUTH_DEFAULT_TITLE",
    "API_AUTH_DEFAULT_VERSION",
    "API_AUTH_DEFAULT_DESCRIPTION",
    "RestMethod",
    "AuthClientType",
    "AuthAPIRoles",
    "EXTENSION_NAME",
    "EXTENSION_BIND",
    "ENDPOINT_SECURITY_LABEL",
    "string_base64",
    "base64_string",
    "info",
    "debug",
    "error",
    "warning",
    "critical",
    "generate_key",
    "generate_salt",
    "hash_string",
    "safe_str_cmp"
]



""" auth_guard module """
import base64
import json
from typing import List, Union
from typing import Tuple
from functools import wraps
from http import HTTPStatus

from flask import request, g, current_app, Response

from .utils.utils import error, debug, warning
from .utils.constants import AuthAPIRoles, ENDPOINT_SECURITY_LABEL

from .internals import get_jwt_extension
from .api.common import response_api_error
from .exceptions import HttpResponseError, ClientAuthenticationError, \
    ClientBadRequestException, ClientIssuerException

from .config import config


def get_token_from_header(allow_missing: bool = False) -> Union[str, None]:
    """
    Get token from header
    """

    token = request.headers.get("Authorization")
    if not token:
        if not allow_missing:
            raise ClientBadRequestException("Authorization header missing")
    else:
        token = str(token).strip()
        if not token.startswith("Basic"):
            # Split token and authentication method
            token_split = token.split(" ")
            auth_method = None
            if len(token_split) > 1:
                auth_method = token_split[0]
                token = token_split[1]
            else:
                # Clean Bearer or Basic label if the user doesn't include the space character
                token = token.replace("Bearer", "").replace("Basic", "")

            if not allow_missing and len(token) == 0:
                raise ClientBadRequestException("Bearer access token missing")

            if auth_method != "Bearer":
                # Other authentication method
                if not allow_missing:
                    raise ClientBadRequestException("Authorization header must follow pattern \"Bearer <token_value>\"")
        else:
            raise ClientBadRequestException("Basic authorization is not allowed in JWT")
        # Set token to None if it is empty string
        token = token if len(token) > 0 else None
    return token


def check_token_format(token: Union[str, None]) -> None:
    """
    Extracts the key id from the token header
    """
    try:
        header = token.split(".")[0]
        # Correct the padding
        header += "=" * (4 - len(header) % 4)
        json.loads(base64.b64decode(header).decode("utf-8"))
    except Exception as e:
        raise ClientBadRequestException("Invalid token format") from e


def verify_jwt(optional: bool = False) -> Tuple[dict, dict]:
    """
    Verify that a valid JWT is present in the request, unless ``optional=True`` in
    which case no JWT is also considered valid.

    :param bool optional:
        If ``True``, do not raise an error if no JWT is present in the request.
        Defaults to ``False``.

    :return:
        A tuple containing the jwt_header and the jwt_data if a valid JWT is
        present in the request. If ``optional=True`` and no JWT is in the request,
        empty dictionaries will be returned instead. Raise an exception if an invalid JWT
        is in the request.
"""

    jwt_data = g._jwt_oceana_jwt_data if hasattr(g, "_jwt_oceana_jwt_data") else {}
    jwt_header = g._jwt_oceana_jwt_header if hasattr(g, "_jwt_oceana_jwt_header") else {}

    raised_exception = None
    try:
        # Get token from header
        token = get_token_from_header(allow_missing=False)
        check_token_format(token=token)

        jwt_extension = get_jwt_extension()
        jwt_data, jwt_header = jwt_extension._decode_jwt_from_config(token=token)
        # Save these at the very end so that they are only saved in the request
        # context if the token is valid and all callbacks succeed.
        g._jwt_oceana_jwt_data = jwt_data
        g._jwt_oceana_jwt_header = jwt_header

    except Exception as e:
        raised_exception = e

    if not optional and raised_exception:
        raise raised_exception
    # Return jwt data and header information
    return jwt_data, jwt_header


def decode_unverified_jwt_request() -> None:
    """
    Decode unverified JWT request
    """
    jwt_data = {}
    unverified_header = {}

    # Get token from header and check
    token = get_token_from_header(allow_missing=True)
    if token:
        try:
            check_token_format(token=token)
            # Decode jwt from request
            jwt_extension = get_jwt_extension()
            jwt_data, unverified_header = jwt_extension._decode_jwt_unverified_from_config(token=token)
        except Exception:
            pass

    # Save jwt in context for public endpoints
    g._jwt_oceana_jwt_data = jwt_data
    g._jwt_oceana_jwt_header = unverified_header

    return jwt_data, unverified_header


def handle_secured_route(endpoint_id, admin, optional, allowed: List[str]) -> bool:
    """
    Handle secure route
    """

    # Verify jwt, if not optional then raises verification exceptions issuer, subject and
    # audience (iss, sub and aud), other custom validations must be set if desired
    jwt_data, jwt_header = verify_jwt(optional=optional)

    # Get roles from jwt
    roles: List[str] = jwt_data.get("roles") or []

    allow = False
    if admin:
        if AuthAPIRoles.ADMIN.value in roles:
            allow = True
        else:
            # Admin can access to the endpoint
            raise ClientAuthenticationError("Only an admin can access to this route")
    else:
        # Verification for token. Will throw HttpResponseError exception if not allowed
        verification_for_token(
            endpoint_id=endpoint_id,
            jwt_header=jwt_header,
            jwt_data=jwt_data,
            optional=optional,
            allowed=allowed,
            roles=roles)
        # If no exception was thrown access is allowed
        allow = True
    return allow


def handle_route_exceptions(route_function, secured, admin, optional, *args, **kwargs):
    """
    Route decorator logic to be executed in a decorated view function.
    Throw exceptions
    """
    # Application configuration contains endpoint security
    endpoint_security = current_app.config[ENDPOINT_SECURITY_LABEL]

    # Get endpoint qualified name
    endpoint_id = route_function.__qualname__

    # Decode JWT token if it is present, and stores jwt and header in
    # user's request variables so can be used afterwards
    jwt_data_unverified, _ = decode_unverified_jwt_request()

    # jwt_extension = get_jwt_extension()
    # aaa = jwt_extension._config().api_secured

    # Api security disabled globally (not recommended)
    if not config.api_secured:
        if jwt_data_unverified is not None:
            client_id = jwt_data_unverified.get(config.identity_claim_key, "unknown") \
                if len(jwt_data_unverified) > 0 else "unknown"
            debug(f"Endpoint \"{endpoint_id}\". Authorization for client_id: \"{client_id}\"")
        return current_app.ensure_sync(route_function)(*args, **kwargs)

    # if endpoint not configured then consider it open as default option
    if endpoint_id not in endpoint_security.keys():
        # Proceed to original route function
        if secured:
            warning(f"Endpoint \"{endpoint_id}\" not configured in security repository")

        if admin:
            # Verify jwt against key
            jwt_data, _jwt_header = verify_jwt(optional=False)
            # Get roles
            roles = jwt_data.get("roles") or []
            if AuthAPIRoles.ADMIN.value not in roles:
                # Other roles distinct that admin can't access to the endpoint
                raise ClientAuthenticationError("Only an admin can access to this route")
        return current_app.ensure_sync(route_function)(*args, **kwargs)

    # Check endpoint roles
    allowed = endpoint_security[endpoint_id].get("roles") or []
    debug(f"Endpoint \"{endpoint_id}\" configured in security repository. Allowed roles: {allowed}")

    # Check JWT roles to determine if the client is authorized.
    # It will throw an exception if it is not authorized
    if secured or admin:
        handle_secured_route(endpoint_id, admin, optional, allowed)

    # Proceed to original route function
    return current_app.ensure_sync(route_function)(*args, **kwargs)


def handle_route(route_function, secured, admin, optional, *args, **kwargs) -> Response:
    """
    Route decorator logic to be executed in a decorated view function.
    Captures exceptions
    """

    # Get endpoint qualified name
    endpoint_id = route_function.__qualname__
    try:
        return handle_route_exceptions(
            route_function,
            secured,
            admin,
            optional,
            *args,
            **kwargs
        )
    except HttpResponseError as e:
        error_msg = f"Bearer {e.error_description()}"
        error(f"{error_msg}")
        http_code = e.status_code
        headers = {"WWW-Authenticate": f"{error_msg}"}
        return response_api_error(http_code=http_code, error=error_msg, headers=headers, endpoint=endpoint_id)
    except Exception as e:
        error_msg = f"{e}"
        http_code = int(HTTPStatus.INTERNAL_SERVER_ERROR.value)
        headers = {"WWW-Authenticate": f"{error_msg}"}
        return response_api_error(http_code=http_code, error=error_msg, headers=headers, endpoint=endpoint_id)
    except BaseException as e:
        error_token = "Token validation failed"
        error_msg = f"Bearer error=\"invalid_token\" error_description=\"{error_token}\""
        error(f"{error_msg}. Exception: {e}")
        http_code = int(HTTPStatus.UNAUTHORIZED.value)
        headers = {"WWW-Authenticate": f"{error_msg}"}
        return response_api_error(http_code=http_code, error=error_msg, headers=headers, endpoint=endpoint_id)


def auth_guard(**kwargs):
    """
    Decorator to protect routes.

    :param bool secured:
        Secured means if client must be valid authenticated with a valid jwt token
    :param bool admin:
        If only role admin is allowed to access endpoint. It overrides endpoint configuration
        from database.
    :param bool optional:
        If ``True``, do not raise an error if no JWT is present in the request.
        Defaults to ``False``.

    :raises:
        :class:`~Exception` if the secret doesn't exist, missing access token or
        Authorization header doesn't follow pattern \"Bearer <token_value>\"
    """
    secured: bool = kwargs.get("secured", False)
    admin: bool = kwargs.get("admin", False)
    optional: bool = kwargs.get("optional", False)

    assert isinstance(secured, bool), "Secured parameter has to be boolean"
    assert isinstance(admin, bool), "Admin parameter has to be boolean"
    assert isinstance(optional, bool), "Optional parameter has to be boolean"

    # Set these properties when admin is set
    if admin:
        secured = True
        optional = False

    def wrapper(route_function):

        @wraps(route_function)
        def decorated_function(*args, **kwargs):
            return handle_route(
                route_function,
                secured,
                admin,
                optional,
                *args,
                **kwargs
            )

        decorated_function.__name__ = route_function.__name__
        return decorated_function
    return wrapper


def get_jwt() -> dict:
    """
    Return python dictionary which has the payload of the JWT that is accessing the endpoint.
    If no JWT is present due to ``auth_guard()``, an empty dictionary is returned.

    :return:
        The payload of the JWT in the current request
    """
    decoded_jwt = g.get("_jwt_oceana_jwt_data", None)
    if decoded_jwt is None:
        raise RuntimeError("You must call `@auth_guard()` before using this method")
    return decoded_jwt


def get_identity() -> str:
    """
    Return the identity of the JWT that is accessing the endpoint.

    :return:
        The identity of the JWT in the current request
    """
    return get_jwt().get(config.identity_claim_key, None)


def verification_for_token(endpoint_id: str,
                           jwt_header: dict,
                           jwt_data: dict,
                           optional: bool,
                           allowed: List[str],
                           roles: List[str]):

    jwt_extension = get_jwt_extension()

    # Issuer only has to be checked when configured jwt as optional, otherwise
    # it has been verified previously
    if optional:
        issuer = jwt_data.get("iss")
        if issuer is not None and issuer != jwt_extension.config().decode_issuer:
            error(f"Issuer error \"{issuer}\" in optional JWT")
            raise ClientIssuerException("Invalid access token: Invalid issuer")

    # Call to token verification function callback
    # Will throw HttpResponseError exception if not allowed
    jwt_extension._token_verification_callback(
        endpoint_id,
        jwt_header,
        jwt_data,
        optional,
        allowed,
        roles)



from sqlalchemy import select
from ..database import db
from ..models import SecIdentity, SecEndpoint


def get_identity(provider, client_type, client_id):
    """
    Get identity information.
    """
    stmt = (
        select(SecIdentity)
        .where(SecIdentity.provider == provider)
        .where(SecIdentity.client_type == client_type)
        .where(SecIdentity.client_id == client_id)
        .where(SecIdentity.enabled == True)  # noqa: E712
    )
    return db.session.execute(statement=stmt).one_or_none()  # .scalars().all()


def get_endpoint_security(provider):

    stmt = (
        select(SecEndpoint).where(SecEndpoint.provider == provider)
    )
    return db.session.execute(statement=stmt).scalars().all()


def get_endpoint_security_dict(provider):
    endpoints = get_endpoint_security(provider)
    return {e.endpoint: e.to_dict() for e in endpoints}



from .models import bearer_token_model


token_doc = {
    "params": {
        "body": {
            "description": "JSON data for the api",
            "in": "body",
            "type": "schema",
            "required": True,
            "schema": {
                "required": ["client_id", "client_secret"],
                "type": "object",
                "properties": {
                    "client_id": {"type": "string", "description": "Client identifier"},
                    "client_secret": {"type": "string", "description": "Client secret"}
                }
            }
        }
    },
    "description": "Authentication",
    "responses": {
        200: ("Created OK", bearer_token_model),
        # 200: ("Created OK"),
        400: "Error",
        401: "Unauthorized",
        403: "Forbidden",
        500: "Internal Server Error"
    }
}



from sqlalchemy import Column, PrimaryKeyConstraint, inspect

from ...database import db
from ...config import Config  # noqa: F401
from ...utils import EXTENSION_BIND

# CREATE TABLE frn.sec_identity (
# 	provider varchar(200) NOT NULL,
# 	client_type varchar(200) NOT NULL,
# 	client_id varchar(1000) NOT NULL,
# 	client_hash varchar(1000) NOT NULL,
# 	client_salt varchar(200) NOT NULL,
# 	roles varchar(1000) NULL,
#   enabled boolean default TRUE,
# 	CONSTRAINT sec_identity_pkey PRIMARY KEY (provider, client_type, client_id)
# );

# ALTER TABLE frn.sec_identity ADD COLUMN enabled BOOLEAN DEFAULT true;


class SecIdentity(db.Model):
    """
    Data model for Identity

    Attributes:
        provider (srt): Provider name
        client_type (str): Client type
        client_id (str): Client ID
        client_hash (str): Client hash
        client_salt (str): Client salt
        roles (str): Roles given to the provider
        enabled (bool): Client enabled
    """
    __bind_key__ = EXTENSION_BIND
    __tablename__ = "sec_identity"
    __table_args__ = (
        PrimaryKeyConstraint("provider", "client_type", "client_id"),
        {
            "extend_existing": True,
            "schema": Config.OCEANA_API_DB_AUTH_SCHEMA
        }
    )

    provider = db.Column(db.String(200), primary_key=True)
    client_type = db.Column(db.String(200), primary_key=True)
    client_id = db.Column(db.String(1000), primary_key=True)
    client_hash = db.Column(db.String(1000), nullable=False)
    client_salt = db.Column(db.String(200), nullable=False)
    roles = db.Column(db.String(1000), nullable=True)
    enabled = db.Column(db.Boolean, default=True)

    def __init__(self, provider, client_type, client_id, client_hash, client_salt, roles, enabled):
        self.provider = provider
        self.client_type = client_type
        self.client_id = client_id
        self.client_hash = client_hash
        self.client_salt = client_salt
        self.roles = roles
        self.enabled = enabled

    def to_dict(self):  # pragma: no cover
        # Convert roles to dictionary
        return {c.key: getattr(self, c.key) if c.key != "roles" else
                [r.strip().lower() for r in str(getattr(self, c.key)).strip().split(",")]
                for c in inspect(self).mapper.column_attrs}

    def __str__(self):  # pragma: no cover
        return ", ".join([f"{c.key}: {getattr(self, c.key)}" for c in inspect(self).mapper.column_attrs])


# CREATE TABLE frn.sec_endpoint (
#     provider varchar(200) NOT NULL,
#     endpoint varchar(200) NOT NULL,
#     roles varchar(1000) NOT NULL,
#     url_template varchar(1000) NULL,
#     description varchar(1000) NULL,
#     CONSTRAINT sec_endpoint_pkey PRIMARY KEY (provider, endpoint)
# );

class SecEndpoint(db.Model):
    __bind_key__ = EXTENSION_BIND
    __tablename__ = "sec_endpoint"
    __table_args__ = (
        PrimaryKeyConstraint("provider", "endpoint"),
        {
            "extend_existing": True,
            "schema": Config.OCEANA_API_DB_AUTH_SCHEMA
        }
    )

    provider = Column(db.String(200), primary_key=True)
    endpoint = Column(db.String(200), primary_key=True)
    roles = Column(db.String(1000), nullable=False)
    url_template = Column(db.String(1000), nullable=True)
    description = Column(db.String(1000), nullable=True)

    def __init__(self, provider, endpoint, roles, url_template, description):
        self.provider = provider
        self.endpoint = endpoint
        self.roles = roles
        self.url_template = url_template
        self.description = description

    def to_dict(self):
        # Convert roles to dictionary
        return {c.key: getattr(self, c.key) if c.key != "roles" else
                [r.strip().lower() for r in str(getattr(self, c.key)).strip().split(",")]
                for c in inspect(self).mapper.column_attrs}

    def __str__(self):  # pragma: no cover
        return ", ".join([f"{c.key}: {getattr(self, c.key)}" for c in inspect(self).mapper.column_attrs])



import os
import logging
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path

from ..config import OCEANA_API_LOGGING_DIR, OCEANA_API_LOGGING_FILE, OCEANA_API_LOGGING_WHEN, \
    OCEANA_API_LOGGING_INTERVAL, OCEANA_API_LOGGING_TITLE, OCEANA_API_LOGGING_LEVEL, \
    OCEANA_API_LOGGING_HANDLERS, OCEANA_API_LOGGING_FORMATTER


class AppLogger:
    """
    A singleton class for application logging.

    This class configures a logger to write logs to a file and console with specified formatting.
    It uses a TimedRotatingFileHandler for log files to rotate them at midnight and keeps 7 days of backup.
    The logger is configured to be a singleton to ensure that only one instance is used throughout the application.
    """
    _instance = None

    def __new__(cls, *args, **kwargs):
        """
        Ensures that only one instance of the class is created.
        """
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, name):
        """
        Initializes the logger with file and console handlers if not already initialized.

        Args:
            - name: The name of the logger, typically the __name__ of the module creating the logger.
        """
        # Prevent re-initialization if already initialized
        if hasattr(self, "logger"):
            return
        # Configuration settings from the config (YAML file)
        # Config handlers
        self._config_handlers()

        # Create logger
        self.logger = logging.getLogger(name)
        self.logger.propagate = False

        if not self.logger.handlers and len(self.log_handlers) > 0:
            level = getattr(logging, OCEANA_API_LOGGING_LEVEL.upper(), logging.INFO)
            self.logger.setLevel(level)

            # "%(asctime)s - [%(name)s] - %(levelname)-5s - %(message)s"
            logFormatter = logging.Formatter(
                OCEANA_API_LOGGING_FORMATTER
            )

            # Add file handler if it is configured
            if "file_handler" in self.log_handlers:
                self.log_dir = OCEANA_API_LOGGING_DIR
                self.log_file = OCEANA_API_LOGGING_FILE

                Path(self.log_dir).mkdir(parents=True, exist_ok=True)
                file_path = os.path.join(self.log_dir, self.log_file)
                fileHandler = TimedRotatingFileHandler(
                    file_path,
                    when=OCEANA_API_LOGGING_WHEN,
                    interval=OCEANA_API_LOGGING_INTERVAL,
                )
                fileHandler.suffix = "%Y%m%d_%H%M%S.log"
                fileHandler.setFormatter(logFormatter)

                self.logger.addHandler(fileHandler)

            if "console" in self.log_handlers:
                consoleHandler = logging.StreamHandler()
                consoleHandler.setFormatter(logFormatter)
                self.logger.addHandler(consoleHandler)

    def _config_handlers(self):
        # Create list of handlers
        _handlers = str(OCEANA_API_LOGGING_HANDLERS)
        self.log_handlers = [
            r.strip().lower() for r in _handlers.strip().split(",")
        ] if len(_handlers) > 0 else []

    def debug(self, message):
        self.logger.debug(message)

    def info(self, message):
        self.logger.info(message)

    def warning(self, message):
        self.logger.warning(message)

    def error(self, message):
        self.logger.error(message)

    def critical(self, message):
        self.logger.critical(message)


logger = AppLogger(OCEANA_API_LOGGING_TITLE)



from .config import BaseConfig, Config, ConfigSqlAlchemy, \
    ConfigPostgres, ConfigSqlite, OCEANA_API_PROVIDER
from .utils import logger, EXTENSION_NAME, EXTENSION_BIND, \
    ENDPOINT_SECURITY_LABEL, API_AUTH_DESCRIPTION, \
    RestMethod, AuthClientType, AuthAPIRoles, \
    string_base64, base64_string, info, debug, error, warning, critical
from .utils.crypto import generate_key, generate_salt, hash_string, safe_str_cmp
from .auth_guard import auth_guard, get_jwt, get_identity, verify_jwt
from .auth import authorizations, security, register_auth_namespace
from .jwt_extension import JWTExtension

from .database import db, get_endpoint_security_dict
from .models import SecIdentity, SecEndpoint
from .jwt_handler import create_access_token

from .auth_provider import authenticate
from .api.common import handle_exceptions

__version__ = "0.0.7"

__all__ = [
    "BaseConfig", "Config", "ConfigSqlAlchemy", "ConfigSqlite", "ConfigPostgres",
    "OCEANA_API_PROVIDER",
    "logger",
    "EXTENSION_NAME",
    "EXTENSION_BIND",
    "ENDPOINT_SECURITY_LABEL",
    "API_AUTH_DESCRIPTION", "RestMethod", "AuthClientType", "AuthAPIRoles",
    "string_base64", "base64_string", "info", "debug", "error", "warning", "critical",
    "generate_key", "generate_salt", "hash_string", "safe_str_cmp",
    "authorizations", "security", "register_auth_namespace",
    "auth_guard", "get_jwt", "get_identity", "verify_jwt",
    "JWTExtension", "db", "get_endpoint_security_dict", "SecIdentity", "SecEndpoint",
    "set_token_verification_function",
    "create_access_token", "authenticate", "handle_exceptions"
]



import os
from setuptools import setup, find_namespace_packages

import pathlib
HERE = pathlib.Path(__file__).parent


PACKAGE_NAME = "oceana_jwt_auth"
VERSION = "0.0.7"
AUTHOR = "jorgegilramos"
DESCRIPTION = "Oceana API library to manage JWT token in Flask Restx applications"
LONG_DESCRIPTION = (HERE / "README.md").read_text(encoding="utf-8")
LONG_DESC_TYPE = "text/markdown"
PROGRAMMING = "Programming Language :: Python :: 3"
LICENSE = "License :: OSI Approved :: MIT License"
OSINFO = "Operating System :: OS Independent"


def parse_requirements(filename):

    filename = f"{HERE}{os.path.sep}{filename}"
    print(f"Filename: {filename}")
    lines_iterator = (line.strip() for line in open(filename))
    return [line for line in lines_iterator if line and not (line.startswith("#") or line.startswith("-"))]


setup(
    name=PACKAGE_NAME,
    version=VERSION,
    packages=find_namespace_packages(where="src", exclude="tests"),
    package_dir={"": "src"},
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESC_TYPE,
    author=AUTHOR,
    license=LICENSE,
    install_requires=parse_requirements("requirements.txt"),
    extras_require={
        "develop": parse_requirements("requirements_test.txt"),
    },
    include_package_data=True,
    # tests_require=parse_requirements("requirements_test.txt")
    classifiers=[
        PROGRAMMING,
        LICENSE,
        OSINFO,
    ],
    python_requires=">= 3.9",
)



from datetime import datetime, timedelta
import uuid
from typing import Dict, Tuple, Union, Any, Iterable, List, Optional
import copy

import jwt
from jwt import ExpiredSignatureError, InvalidSignatureError, \
    DecodeError, InvalidTokenError

from .internals import get_jwt_extension
from .exceptions import ClientJWTDecodeException, ClientAuthenticationError, \
    ClientBadRequestException, ClientInvalidSignatureException


def encode_jwt(
    algorithm: str,
    audience: Union[str, Iterable[str]],
    claim_overrides: dict,
    expires_delta: timedelta,
    header_overrides: dict,
    identity: Any,
    identity_claim_key: str,
    issuer: str,
    roles: List[str],
    secret: str,
    token_type: str,
    nbf: bool,
    version: str = None,
    payload: Dict = {}

) -> Tuple[str, Dict]:
    """
    Generates a new JWT token with the provided information
    """
    _now = datetime.now()

    token_data = copy.deepcopy(payload)
    token_data.update({
        "iat": _now.timestamp(),
        "jti": str(uuid.uuid4()),
        "type": token_type,
        identity_claim_key: identity,
        "roles": roles,
        # Added "created" in human format date
        "created": _now.strftime("%Y-%m-%d %H:%M:%S")
    })

    if nbf:
        token_data["nbf"] = _now.timestamp()

    if audience:
        token_data["aud"] = audience

    if issuer:
        token_data["iss"] = issuer

    if expires_delta:
        valid_until = _now + expires_delta
        token_data["exp"] = valid_until.timestamp()
        # Added "expires_in" in human format date
        token_data["expires_in"] = valid_until.strftime("%Y-%m-%d %H:%M:%S")

    if version:
        token_data["version"] = version

    if claim_overrides:
        token_data.update(claim_overrides)

    return jwt.encode(
        payload=token_data,
        key=secret,
        algorithm=algorithm,
        headers=header_overrides
    ), token_data


def _decode_jwt(token: str,
                key: str,
                algorithms: List,
                audience: Union[str, Iterable[str]],
                issuer: str,
                options: Dict[str, Any]) -> Tuple[Dict, Dict]:
    """
    Tries to retrieve payload information inside of a existent JWT token (string)
    Will throw an error if the token is invalid (expired or inconsistent)
    """

    try:
        # Call to pyjwt library to verify claims.
        # Always verify ext, iat, and nbf claims, and exp and aud if enabled.
        header = jwt.get_unverified_header(token)
        payload = jwt.decode(
            jwt=token,
            key=key,
            algorithms=algorithms,
            audience=audience,
            issuer=issuer,
            options=options,
        )
    except ExpiredSignatureError as e:
        raise ClientAuthenticationError(f"Invalid access token: {e}")  # 401
    except (InvalidSignatureError, DecodeError) as e:
        raise ClientInvalidSignatureException(f"Invalid access token: {e}")  # 400
    except InvalidTokenError as e:
        raise ClientAuthenticationError(f"Invalid access token: {e}")  # 401
    # Raise rest of exceptions
    return payload, header


def decode_jwt(
    algorithms: List,
    audience: Union[str, Iterable[str]],
    token: str,
    identity_claim_key: str,
    issuer: str,
    key: str,
    version: str,
    verify_aud: bool,
    verify_sub: bool,
    verify_version: bool
) -> Tuple[Dict, Dict]:

    options = {"verify_aud": verify_aud, "verify_sub": verify_sub}

    payload, header = _decode_jwt(token=token,
                                  key=key,
                                  algorithms=algorithms,
                                  audience=audience,
                                  issuer=issuer,
                                  options=options)

    # Make sure that any custom claims we expect in the token are present
    if identity_claim_key not in payload:
        raise ClientJWTDecodeException(f"Invalid access token: Missing claim: {identity_claim_key}")

    if verify_version:
        if "version" not in payload:
            raise ClientJWTDecodeException("Invalid access token: Missing claim: version")

        if payload["version"] != version:
            raise ClientJWTDecodeException("Invalid access token: Token version doesn't match")

    # Set default properties if not present
    if "type" not in payload:
        payload["type"] = "access"

    if "jti" not in payload:
        payload["jti"] = None

    return payload, header


def decode_jwt_unverified(
    algorithms: List,
    token: str,
) -> Tuple[Dict, Dict]:

    try:
        header = jwt.get_unverified_header(token)
        unverified_payload = jwt.decode(
            token,
            algorithms=algorithms,
            # Also doesn't raise expiration signature exception
            options={"verify_signature": False}
        )
    except Exception as e:
        raise ClientBadRequestException(f"Invalid access token: {e}")

    return unverified_payload, header


def create_access_token(
    identity: Any,
    roles: List[str],
    expires_delta: Optional[timedelta] = None,
    additional_claims=None,
    additional_headers=None,
    payload: Dict = {}
) -> Tuple[str, Dict]:

    # Get extension
    jwt_manager = get_jwt_extension()
    return jwt_manager._encode_jwt_from_config(
        identity=identity,
        roles=roles,
        token_type="access",
        claims=additional_claims,
        expires_delta=expires_delta,
        headers=additional_headers,
        version=None,
        payload=payload
    )



import os
import base64
import hashlib
import hmac


def generate_key() -> bytes:
    """
    Generate key in bytes
    """
    return base64.urlsafe_b64encode(os.urandom(32))


def generate_salt() -> str:
    """
    Generate salt key url safe
    """
    return generate_key().decode("utf-8")


def hash_string(source_str: str, salt: str) -> str:
    """
    Create a hash using SHA-256 algorithm
    """

    # Create a hash object using SHA-256 algorithm
    hash_object = hashlib.sha256()

    # Update the hash object with a message
    hash_object.update(str(salt).encode("utf-8") + str(source_str).encode("utf-8"))

    # Get the hexadecimal digest of the hash
    hex_dig = hash_object.hexdigest()

    return hex_dig


def safe_str_cmp(a: str, b: str) -> bool:
    """
    This function compares strings in constant time.

    Returns `True` if the two strings are equal, or `False` if they are not.
    """

    if isinstance(a, str):
        a = a.encode("utf-8")  # type: ignore

    if isinstance(b, str):
        b = b.encode("utf-8")  # type: ignore

    return hmac.compare_digest(a, b)



"""
JWT Extension
"""
from datetime import timedelta
from typing import Optional, Callable, Any, List, Tuple, Dict, Union

from flask import Blueprint, Flask
from flask_sqlalchemy import SQLAlchemy
from flask_restx import Api
# from flask_wtf.csrf import CSRFProtect
from urllib.parse import quote_plus

from .config import Config, OCEANA_API_PROVIDER
from .utils import info, debug, EXTENSION_NAME, EXTENSION_BIND, ENDPOINT_SECURITY_LABEL, \
    API_AUTH_DEFAULT_TITLE, API_AUTH_DEFAULT_VERSION, API_AUTH_DEFAULT_DESCRIPTION
from .internals import default_user_claims_callback, default_token_header_callback, \
    default_token_verification_callback, default_encode_key_callback, \
    default_decode_key_callback
from .auth import authorizations, security, register_auth_namespace
from .database import db, init_app, get_endpoint_security_dict

from .jwt_handler import encode_jwt, decode_jwt, decode_jwt_unverified


class JWTExtension():
    """
    JWTExtension
    """

    def __init__(self,
                 app: Optional[Flask] = None,
                 api: Optional[Api] = None,
                 config_object: Optional[Callable] = None,
                 declarative_base: SQLAlchemy = None,
                 *args,
                 **kwargs):
        """
        Create an jwt authorization helper for a application.

        :param Flask app:
            It can be a Flask app instance, or ``None``.

            1. Use your globally avalaible app object.

            2. But if you are using `Application Factory pattern
            <https://flask.palletsprojects.com/en/latest/patterns/appfactories/>`,
            your app is not available globally, so you need to pass ``None`` here,
            and call :func:`JWTExtension.init_app()` later,
        :param Api api:
            Api Flask RestX instance, or ``None``.
            If app object is passed, it is mandatory to provide Api object as well

        """

        # Store config object
        self._config = config_object() if config_object is not None else Config()

        # Define callback for verification token function
        self._token_verification_callback = default_token_verification_callback
        self._user_claims_callback = default_user_claims_callback
        self._encode_key_callback = default_encode_key_callback
        self._decode_key_callback = default_decode_key_callback
        self._token_header_callback = default_token_header_callback

        if app is not None:
            self.init_app(app, api, config_object, declarative_base, *args, **kwargs)

    def init_app(self,
                 app: Optional[Flask] = None,
                 api: Optional[Api] = None,
                 config_object: Optional[Callable] = None,
                 declarative_base: SQLAlchemy = None,
                 *args,
                 **kwargs) -> Flask:
        """
        Register this extension with the flask app.

        :param app:
            The Flask Application object
        :param api:
            Flask RestX Api instance
        """

        if app is not None and EXTENSION_NAME in app.extensions:
            raise RuntimeError(
                f"A \"{EXTENSION_NAME}\" instance has already been registered on this Flask app."
                " Import and use that instance instead."
            )

        if app is None:
            # Create flask application if it is not provided
            app = Flask(__name__)
            if api is None:
                bp = Blueprint(f"{EXTENSION_NAME}_app", __name__)
                api = self.create_api(bp, **kwargs)
                app.register_blueprint(bp)
        elif api is None:
            raise Exception("Api parameter can't be empty when app is specified")

        # Store config object
        if config_object is not None:
            self._config = config_object()

        # Store Api instance
        self._app = app
        self._api = api

        # Add authorization configuration
        api.authorizations = authorizations
        api.security = security

        app = self._create_app(app, api, config_object, declarative_base, *args, **kwargs)

        # Check extension property
        app.extensions = {} if not hasattr(app, "extensions") else app.extensions
        app.extensions[EXTENSION_NAME] = self
        return app

    def create_api(self, bp=None, **kwargs):
        _title = kwargs.get("title") or API_AUTH_DEFAULT_TITLE
        _version = kwargs.get("version") or API_AUTH_DEFAULT_VERSION
        _description = kwargs.get("description") or API_AUTH_DEFAULT_DESCRIPTION
        api = Api(
            app=bp,
            authorizations=authorizations,
            security=security,
            title=_title,
            version=_version,
            description=_description
        )
        return api

    def _create_app(self,
                    app: Flask,
                    api: Api,
                    config_object: Optional[Callable] = None,
                    declarative_base: SQLAlchemy = None,
                    *args,
                    **kwargs):
        # Default config is Config
        if not config_object:
            config_object = Config
            self._config = Config()

        # Get properties from object. SQLAlchemy properties are taken from
        # the Flask application overriding those ones coming from the object.
        # app.config.from_object(config_object())
        self._app_config_from_object(app, config_object)

        # TODO: Init other extensions
        # csrf.init_app(app)
        # oauth.init_app(app)
        # register_oauth_clients(app)

        # Init database context
        if declarative_base is None:
            declarative_base = db

        # Initialize SQLAlchemy application
        if "sqlalchemy" not in app.extensions:
            declarative_base.init_app(app)

        # self._declarative_base = declarative_base

        with app.app_context():
            testing = hasattr(app, "testing") and bool(app.testing)
            debug(f"Testing: {testing}")

            # Show information about global security
            info(f"API secured: {self._config.api_secured}")

            # Create all necessary database entities
            init_app(config_object, testing, declarative_base)

            # Get endpoint security from database
            self.update_auth_from_db()

            # Register authorization namespace (need context)
            if self._config.register_auth:
                bp = Blueprint(f"{EXTENSION_NAME}_api", __name__, url_prefix="/")
                register_auth_namespace(api=api)
                app.register_blueprint(bp)
            info(f"Registered authorization endpoints: {self._config.register_auth}")

        return app

    def _app_config_from_object(self,
                                app: Flask,
                                config_object: Optional[Callable] = None):
        sqlalchemy_keys = ["SQLALCHEMY_DATABASE_URI",
                           "SQLALCHEMY_ENGINE_OPTIONS",
                           "SQLALCHEMY_ECHO",
                           "SQLALCHEMY_BINDS",
                           "SQLALCHEMY_RECORD_QUERIES",
                           "SQLALCHEMY_TRACK_MODIFICATIONS"]

        obj = config_object()
        for key in dir(obj):
            if key.isupper():
                if key not in sqlalchemy_keys:
                    value = getattr(obj, key)
                else:
                    value = getattr(obj, key) if not app.config.get(key) \
                        else app.config[key]
                app.config[key] = value
                setattr(self._config, key, value)

        # Ensure oceana_jwt_auth bind
        uri = self._escape_uri(
            uri=app.config["SQLALCHEMY_DATABASE_URI"],
            passwd=config_object.OCEANA_API_DB_AUTH_PASSWORD)

        if (_binds := app.config.get("SQLALCHEMY_BINDS")) is None:
            app.config["SQLALCHEMY_BINDS"] = \
                {EXTENSION_BIND: uri}
        elif EXTENSION_BIND not in _binds:
            app.config["SQLALCHEMY_BINDS"].update(
                {EXTENSION_BIND: uri}
            )
        else:
            uri = self._escape_uri(
                uri=app.config["SQLALCHEMY_BINDS"][EXTENSION_BIND],
                passwd=config_object.OCEANA_API_DB_AUTH_PASSWORD)
            app.config["SQLALCHEMY_BINDS"][EXTENSION_BIND] = uri

    def _escape_uri(self, uri: str, passwd: str) -> str:

        passwd_in_uri = str(uri.split("://")[1].split("@")[0]).find(":") > 0
        db_auth_uri = uri
        if not passwd_in_uri:
            if passwd is not None:
                uri_split = uri.split("@")
                if len(uri_split) == 2:
                    db_auth_uri = "".join([
                        uri_split[0],
                        ":",
                        quote_plus(passwd),  # Escape passwd
                        "@",
                        uri_split[1]])
                else:
                    raise RuntimeError("Syntax mismatch in database URI")
        return db_auth_uri

    def update_auth_from_db(self):

        # Get endpoint security from database
        secured_endpoints = get_endpoint_security_dict(provider=OCEANA_API_PROVIDER)
        self._app.config[ENDPOINT_SECURITY_LABEL] = secured_endpoints
        setattr(self._app.config, ENDPOINT_SECURITY_LABEL, secured_endpoints)
        info(f"Secured endpoints: {len(secured_endpoints)}")
        # Show security endpoints information from database when the application is started
        for endpoint_id in secured_endpoints:  # pragma: no cover
            roles = secured_endpoints[endpoint_id].get("roles")
            info(f"    - {endpoint_id}: {roles}")
        debug(f"Endpoint security: {secured_endpoints}")

    def config(self) -> Config:
        """
        Get extension configuration instance
        """
        return self._config

    def api(self) -> Api:
        """
        Get application Api instance (Flask RestX)
        """
        return self._api

    # Decorated functions to override default behaviour
    def set_token_verification_function(self, callback: Callable) -> Callable:
        """
        Decorator function. Sets the callback function used for custom
        verification of a valid JWT.

        Decorated function:
            Must take six arguments.

        :param str endpoint_id: Argument indicating the endpoint identificator.
        :param dict jwt_header: Dictionary containing the header data of the JWT.
        :param dict jwt_data: Dictionary containing the payload data of the JWT.
        :param bool optional: Boolean ``True`` if JWT is optional in the endpoint
            ``False`` otherwise.
        :param list allowed: List of roles allowed for the endpoint.
        :param list roles: List of roles from the valid JWT

        :return None:
        """

        self._token_verification_callback = callback
        return callback

    def set_user_claims_function(self, callback: Callable) -> Callable:
        """
        Decorator function. Sets the callback function used to add custom
        user claims in a JWT.
        The dictionary of claims returned in the decorated function will be
        added to claims in JWT that were passed in parameter in ``additional_claims``
        when creating the access token. Parameter ``additional_claims`` override
        claims in case of name coincidence.

        Decorated function:
            Must take only one argument.

        :param str identity:
            Argument indicating the endpoint identificator.

        :return dict user_claims:
            Dictionary with custom user claims in a JWT.
        """

        self._user_claims_callback = callback
        return callback

    def set_token_header_function(self, callback: Callable) -> Callable:
        """
        Decorator function. Sets the callback function used to add custom user
        properties to the header in a JWT.

        Decorated function:
            Must take only one argument.

        :param str header: Argument indicating headers to add.

        :return dict header_properties:
            Dictionary with custom properties to the header in a JWT.
        """

        self._token_header_callback = callback
        return callback

    def _encode_jwt_from_config(
        self,
        identity: Any,
        roles: List[str],
        token_type: str,
        claims=None,
        expires_delta: Optional[timedelta] = None,
        headers=None,
        version=None,
        payload: dict = {}
    ) -> Tuple[str, Dict]:
        # Override headers
        add_headers = self._token_header_callback(headers)
        if headers is not None:
            add_headers.update(headers)

        # Override claims
        add_claims = self._user_claims_callback(identity)
        if claims is not None:
            add_claims.update(claims)

        config = self._config  # self.config()
        if expires_delta is None:
            if token_type == "access":
                expires_delta = config.access_token_delta
            # TODO: implement refresh tokens
            # else:
            #     expires_delta = config.refresh_token_delta

        if version is None and config.token_version is not None:
            version = config.token_version

        return encode_jwt(
            algorithm=config.algorithm,
            audience=config.encode_audience,
            claim_overrides=add_claims,
            expires_delta=expires_delta,
            header_overrides=add_headers,
            identity=identity,
            identity_claim_key=config.identity_claim_key,
            issuer=config.encode_issuer,
            roles=roles,
            secret=self._encode_key_callback(identity, config.algorithm),
            token_type=token_type,
            nbf=True,
            version=version,
            payload=payload
        )

    def _decode_jwt_from_config(
        self,
        token: Union[str, None],
        version=None,
    ) -> Tuple[Dict, Dict]:
        """
        Called from verify_jwt_request
        """

        config = self._config
        algorithms = [config.algorithm]
        key = self._decode_key_callback(identity=None, algorithm=config.algorithm)

        if version is None and config.token_version is not None:
            version = config.token_version

        return decode_jwt(
            algorithms=algorithms,
            audience=config.decode_audience,
            token=token,
            identity_claim_key=config.identity_claim_key,
            issuer=config.decode_issuer,
            key=key,
            version=version,
            verify_aud=config.decode_audience is not None,
            verify_sub=config.verify_sub,
            verify_version=config.verify_version)

    def _decode_jwt_unverified_from_config(
        self,
        token: str
    ) -> Tuple[Dict, Dict]:
        """
        Tries to retrieve and header and payload information inside of a existent JWT token (string)
        """
        algorithms = [self._config.algorithm]
        return decode_jwt_unverified(
            algorithms=algorithms,
            token=token
        )



from flask import request
from flask_restx import Namespace, Resource, Api

from ..config import OCEANA_API_PROVIDER, OCEANA_API_AUTH_VERSION
from ..utils import info
from ..utils.constants import RestMethod, API_AUTH_DESCRIPTION
from ..api.common import response_api_ok, handle_exceptions
from ..auth_provider import authenticate
from ..jwt_handler import create_access_token

# Models and documents for swagger
from .models import bearer_token_model
from .docs import token_doc

from ..exceptions import ClientBadRequestException

authorizations = {
    "JWT": {
        "type": "apiKey",
        "in": "header",
        "name": "Authorization"
    }
    # https://swagger.io/docs/specification/v3_0/authentication/oauth2/
    # "entraid": {
    #     "type": "oauth2",
    #     "flow": "accessCode",
    #     "tokenUrl": "https://somewhere.com/token",
    #     "authorizationUrl": "https://somewhere.com/auth",
    #     "scopes": {
    #         "read": "Grant read-only access",
    #         "write": "Grant read-write access",
    #     }
    # }
}

security = "JWT"  # ["JWT", {"entraid": "read"}]


def default_authentication(client_id, client_secret):

    identity = {}
    if not client_id or not client_secret:
        raise ClientBadRequestException("Client id or client secret missing")

    identity = authenticate(provider=OCEANA_API_PROVIDER,
                            client_id=client_id,
                            client_secret=client_secret)
    return identity


def register_auth_namespace(api: Api):

    # Create auth namespace
    ns_auth = Namespace(
        name="Auth",
        description=f"{API_AUTH_DESCRIPTION}",
        path=f"/{OCEANA_API_AUTH_VERSION}/auth" if OCEANA_API_AUTH_VERSION else "/auth")
    # Register model
    ns_auth.model(name="Token", model=bearer_token_model)

    # Add authorization namespace
    api.add_namespace(ns_auth)

    @ns_auth.route("/token", methods=[RestMethod.POST.value])
    @ns_auth.doc(**token_doc)
    class GetToken(Resource):
        @handle_exceptions(endpoint="GetToken.post")
        def post(self):
            endpoint = "GetToken.post"
            info(f"Starting {endpoint} endpoint call ...")

            client_id = request.json.get("client_id")
            client_secret = request.json.get("client_secret")

            identity = default_authentication(client_id=client_id, client_secret=client_secret)

            client_id = identity["client_id"]
            roles = identity["roles"]

            info(f"{endpoint} - client_id: {client_id} - roles: {roles}")

            # Payload that the server will send back the client encoded in the JWT Token.
            # While generating a token, you can define any type of payload in valid JSON format
            # the iss(issuer), sub(subject) and aud(audience) are reserved claims.
            # https://tools.ietf.org/html/rfc7519#section-4.1
            # These reserved claims are not mandatory to define in a standard JWT token.

            token, payload = create_access_token(
                identity=client_id,
                roles=roles,
                expires_delta=None,
                additional_claims=None,
                additional_headers=None,
                payload={}
            )

            # Add token to return
            data = {
                "token": f"{token}"
            }

            # Added label "expiresIn" so applications can request for new tokens before
            # the token expiration time without the need to decode the JWT or receive an
            # https error from the authorization server.
            expires_in = payload.get("expires_in")
            if expires_in is not None:
                data = {**data, "expiresIn": expires_in}

            # Return token
            return response_api_ok(http_code=200, data=data, endpoint=None)
